#include "clientprot.h"
#include "math.h"
#include "player.h"
#include "map.h"
#include "checksum.h"
#include "fileio.h"
#include "zlib.h"
#include "command.h"


//////// Switches ////////

void handleUnknown		(hostMessage * msg)
{
	/*	Field	Length	Description
		0		1		Type byte
		1		...		?
	*/

	printf("Unknown message type %i(%i)\n", msg->msg[0], msg->len);
	WriteLog(msg->msg, msg->len);
}

void handleKeepAlive		(hostMessage * msg)
{
	/*	Field	Length	Description
		0		1		Type byte
	*/

	if (msg->len != 1)
	{
		handleUnknown(msg);

		return;
	}
}

void handleBannerFlag		(hostMessage * msg)
{
	/*	Field	Length	Description
		0		1		Type byte
		1		1		?
	*/

	BYTE toggle = getByte(msg->msg, 1);

	if (toggle || (msg->len != 2))
	{
		handleUnknown(msg);

		return;
	}
}

void handleLoginNext		(hostMessage * msg)
{
	/*	Field	Length	Description
		0		1		Type byte
	*/

	if (msg->len != 1)
	{
		handleUnknown(msg);

		return;
	}
}

void handleInGameFlag		(hostMessage * msg)
{
	/*	Field	Length	Description
		0		1		Type byte
	*/

	if (msg->len != 1)
	{
		handleUnknown(msg);

		return;
	}

	msg->src->position = true;
	msg->src->lastPosition = msg->src->getTime();

	printf("Spawn connected.\n");
}

void handleBannerAds		(hostMessage * msg)
{	// Provided by Ave-iator
	/*	Field	Length	Description
		0		1		Type byte
		1		1		Display mode
		2		2		Width
		4		2		Height
		6		4		Duration
		10		?		Banner indices
	*/

	BYTE mode = getByte(msg->msg, 1);
	Uint16 width = getShort(msg->msg, 2);
	Uint16 height = getShort(msg->msg, 4);
	Uint32 duration = getLong(msg->msg, 6);

	// I'm not sure how these are packed (W*H+10 != msg->len)

	printf("Got banner advertisement(s)\n");
}


//////// FTP ////////

void handleFileTransfer		(hostMessage * msg)
{
	/*	Field	Length	Description
		0		1		Type byte
		1		16		File name
		17		...		Compressed file
	*/

	char * name = msg->msg + 1;
	char * file = msg->msg + 17;
	Uint32 length = msg->len - 17;

	if (name[15])
	{
		printf("WARNING: Unterminated downloaded-file name ignored.\n");

		return;
	}

	for (Uint16 i = 0; i < 16; i++)
	{
		char c = name[i];

		if (c == 0)
			break;

		switch (c)
		{
		case '/':
		case '\\':
			printf("WARNING: Invalid downloaded-file name path ignored. (%s)\n", name);
			return;
		default:
			break;
		};

		if ((c < ' ') || (c > '~'))
		{
			printf("WARNING: Invalid downloaded-file name chars ignored.\n");
			return;
		}
	}

	fileAccess fa;

	if (name[0])
	{
		printf("Got file: %s\n", name);

		if (fa.open(name, false))
		{
			fa.setFile(file, length);
		}
	}
	else
	{
		printf("Got news file.\n");
		strcpy(name, "news.txt");

		if (fa.open(name, false))
		{
			fa.decompress(file, length);
		}
	}
}

void handleFileRequest		(hostMessage * msg)
{
	/*	Field	Length	Description
		0		1		Type byte
		1		256		Local file name
		257		16		Remote file name
	*/

	char * fileName = msg->msg + 1;
	char * remoteName = msg->msg + 257;

	if (msg->len != 273)
	{
		printf("WARNING: Requested-file malformed buffer ignored.\n");
		WriteLog(msg->msg, msg->len);

		return;
	}

	if (fileName[255] || (fileName[0] == 0))
	{
		printf("WARNING: Unterminated/blank requested-file local name ignored.\n");
		WriteLog(msg->msg, msg->len);

		return;
	}

	if (remoteName[15] || (remoteName[0] == 0))
	{
		printf("WARNING: Unterminated/blank requested-file remote name ignored.\n");
		WriteLog(msg->msg, msg->len);

		return;
	}

	for (Uint16 i = 0; i < 16; i++)
	{
		char c = fileName[i];

		if (c == 0)
			break;

		switch (c)
		{
		case '/':
		case '\\':
			printf("WARNING: Invalid requested-file name path ignored. (%s)\n", fileName);
			return;
		default:
			break;
		};

		if ((c < ' ') || (c > '~'))
		{
			printf("WARNING: Invalid requested-file name chars ignored.\n");
			return;
		}
	}

	// Put other restrictions here

	printf("File request: %s\n", fileName);

	// Encode and transmit file here
	fileAccess fa;
	if (fa.open(fileName, true) == false)
	{
		printf("Unable to open file\n");

		return;
	}

	Uint32 len = fa.getLength();
	char * file = fa.readStuff(0, len);

	clientMessage * cm = generateFileTransfer(fileName, file, len);
	if (cm)
	{
		msg->src->post(cm->msg, cm->len, true);
		delete cm;
	}

	delete []file;
}


//////// Turret ////////

void handleCreateTurret		(hostMessage * msg)
{
	/*	Field	Length	Description
		0		1		Type byte
		1		2		Turreter ident
		3		2		Turretee ident
	*/

	if (msg->len != 5)
	{
		handleUnknown(msg);

		return;
	}

	Player * turreter = msg->src->getPlayer(getShort(msg->msg, 1));
	if (turreter == NULL) return;
	Player * turretee = msg->src->getPlayer(getShort(msg->msg, 3));

	printf("Player turreted: %s\n", turreter->name);
}

void handleDeleteTurret		(hostMessage * msg)
{
	/*	Field	Length	Description
		0		1		Type byte
		1		2		Player ident
	*/

	if (msg->len != 3)
	{
		handleUnknown(msg);

		return;
	}

	Player * p = msg->src->getPlayer(getShort(msg->msg, 1));
	if (p == NULL) return;

	printf("Deleted turret: %s\n", p->name);
}


//////// Flags ////////

void handleTurfFlagStatus	(hostMessage * msg)
{
	/*	Field	Length	Description
		0		1		Type byte
	The following are repeated until the end of the message
		1		2		Team for flag X
	*/

	Uint32 index = 1;
	Uint16 team,
		   flag = 0;

	if ((msg->len - 1) & 1)
	{
		handleUnknown(msg);

		return;
	}

	while (index < msg->len)
	{
		team = getShort(msg->msg, index);

		Flag * f = msg->src->findFlag(flag);
		f->team = team;

		index += 2;
		flag++;
	}

	printf("Got turf update\n");
}

void handleFlagPosition		(hostMessage * msg)
{
	/*	Field	Length	Description
		0		1		Type byte
		1		2		Flag ident
		3		2		X tiles
		5		2		Y tiles
		7		2		Team
	*/

	if (msg->len != 9)
	{
		handleUnknown(msg);

		return;
	}

	Uint16 flag = getShort(msg->msg, 1);
	Uint16 x = getShort(msg->msg, 3);
	Uint16 y = getShort(msg->msg, 5);
	Uint16 team = getShort(msg->msg, 7);

	Flag * f = msg->src->findFlag(flag);
	f->x = x;
	f->y = y;
	f->team = team;
}

void handleFlagVictory		(hostMessage * msg)
{
	/*	Field	Length	Description
		0		1		Type byte
		1		2		Team
		3		4		Points
	*/

	if (msg->len != 7)
	{
		handleUnknown(msg);

		return;
	}

	Uint16 team = getShort(msg->msg, 1);
	Uint32 points = getLong(msg->msg, 3);

	if (team == PROT_ERROR)
	{
		// Flag game reset
		printf("Flag game reset.\n");
	}
	else
	{
		// Team victory
		printf("Team # %i won the flag game for %i points.\n", team, points);

		_listnode <Player> * parse = msg->src->playerlist.head;

		while (parse)
		{
			Player * p = parse->item;
			parse = parse->next;

			if (p->team == team)
				p->flagPoints += points;
		}
	}

	msg->src->resetFlags();
}

void handleFlagDrop			(hostMessage * msg)
{
	/*	Field	Length	Description
		0		1		Type byte
		1		2		Player ident
	*/

	if (msg->len != 3)
	{
		handleUnknown(msg);

		return;
	}

	Player * p = msg->src->getPlayer(getShort(msg->msg, 1));
	if (p == NULL) return;

	printf("Dropped flags: %s\n", p->name);
}

void handleFlagClaim		(hostMessage * msg)
{
	/*	Field	Length	Description
		0		1		Type byte
		1		2		Flag ident
		3		2		Player ident
	*/

	if (msg->len != 5)
	{
		handleUnknown(msg);

		return;
	}

	Uint16 flag = getShort(msg->msg, 1);
	Player * p = msg->src->getPlayer(getShort(msg->msg, 3));
	if (p == NULL) return;

	printf("Snagged flag # %i: %s\n", flag, p->name);
}

void handleFlagReward		(hostMessage * msg)
{
	/*	Field	Length	Description
		0		1		Type byte
	The following are repeated until the end of the message
		1		2		Team
		3		2		Points
	*/

	Uint32 index = 1;

	if ((msg->len - 1) & 3)
	{
		handleUnknown(msg);

		return;
	}

	while (index < msg->len)
	{
		Uint16 team = getShort(msg->msg, index);
		Uint32 points = getShort(msg->msg, index + 2);

		printf("Flag reward for team %i: %i\n", team, points);

		_listnode <Player> * parse = msg->src->playerlist.head;

		while (parse)
		{
			Player * p = parse->item;
			parse = parse->next;

			if (p->team == team)
				p->flagPoints += points;
		}

		index += 4;
	}
}


//////// Players ////////

void handleScoreReset		(hostMessage * msg)
{
	/*	Field	Length	Description
		0		1		Type byte
		1		2		Player ident
	*/

	if (msg->len != 3)
	{
		handleUnknown(msg);

		return;
	}

	Uint16 ident = getShort(msg->msg, 1);
	Player * p = msg->src->getPlayer(ident);

	if (ident != PROT_ERROR)
	{
		if (p == NULL) return;

		printf("Score reset: %s\n", p->name);

		p->flagPoints = 0;
		p->killPoints = 0;
		p->wins = 0;
		p->losses = 0;
	}
	else
	{
		printf("Score reset: All\n");

		_listnode <Player> * parse = msg->src->playerlist.head;

		while (parse)
		{
			Player * p = parse->item;
			parse = parse->next;

			p->flagPoints = 0;
			p->killPoints = 0;
			p->wins = 0;
			p->losses = 0;
		}
	}
}

void handlePlayerPrize		(hostMessage * msg)
{
	/*	Field	Length	Description
		0		1		Type byte
		1		4		Timestamp
		5		2		X tiles
		7		2		Y tiles
		9		2		Prize type
		11		2		Player ident
	*/

	if (msg->len != 13)
	{
		handleUnknown(msg);

		return;
	}

	Uint32 timestamp = msg->src->getLocalTime(getLong(msg->msg, 1));
	Uint16 x = getShort(msg->msg, 5);
	Uint16 y = getShort(msg->msg, 7);
	Uint16 prize = getShort(msg->msg, 9);
	Player * p = msg->src->getPlayer(getShort(msg->msg, 11));
	if (p == NULL) return;

//	printf("Player picked up a prize: %s\n", p->name);
}

void handlePlayerEntering	(hostMessage * msg)
{	// Average and rating?
	/*	Field	Length	Description
	The following are repeated until the end of the message
		0		1		Type byte
		1		1		Ship type
		2		1		?
		3		20		Player name
		23		20		Squad name
		43		4		Flag points
		47		4		Kill points
		51		2		Player ident
		53		2		Team
		55		2		Wins
		57		2		Losses
		59		2		Turretee ident
		61		2		Flags carried
		63		1		Boolean: Has KoTH
	*/

	Uint32 index = 0;

	if (msg->len & 63)
	{
		handleUnknown(msg);

		return;
	}

	while (index < msg->len)
	{
		BYTE ship = getByte(msg->msg, index + 1);
		char * name = msg->msg + index + 3;
		char * squad = msg->msg + index + 23;
		Uint32 flagPoints = getLong(msg->msg, index + 43);
		Uint32 killPoints = getLong(msg->msg, index + 47);
		Uint16 ident = getShort(msg->msg, index + 51);
		Uint16 team = getShort(msg->msg, index + 53);
		Uint16 wins = getShort(msg->msg, index + 55);
		Uint16 losses = getShort(msg->msg, index + 57);
		Uint16 turretee = getShort(msg->msg, index + 59);
		Uint16 flagCount = getShort(msg->msg, index + 61);
		bool hasKoTH = getByte(msg->msg, index + 63) ? 1 : 0;

		printf("Player entering: %s [%i]", name, getByte(msg->msg, index + 2));

		if (flagCount)
		{
			printf(" {Flags:%i}", flagCount);
		}

		if (hasKoTH)
		{
			printf(" {KOTH}");
		}

		if (ident == msg->src->Me)
		{
			if (hasKoTH)
			{
				clientMessage * cm = generateKoTHReset();
				msg->src->post(cm->msg, cm->len, true);
				delete cm;
			}

			printf("  <-- Me");
		}

		printf("\n");

		msg->src->addPlayer(ident, name, squad, flagPoints, killPoints, team, wins, losses, ship);

		index += 64;
	}
}

void handleSetTeam			(hostMessage * msg)
{
	/*	Field	Length	Description
		0		1		Type byte
		1		2		Player ident
		3		2		Team
		5		1		?
	*/

	if (msg->len != 6)
	{
		handleUnknown(msg);

		return;
	}

	Player * p = msg->src->getPlayer(getShort(msg->msg, 1));
	if (p == NULL) return;
	Uint16 team = getShort(msg->msg, 3);

	printf("Player changed teams: %s\n", p->name);

	p->team = team;
}

void handleSetTeamAndShip	(hostMessage * msg)
{
	/*	Field	Length	Description
		0		1		Type byte
		1		1		Ship type
		2		2		Player ident
		4		2		Team
	*/

	if (msg->len != 6)
	{
		handleUnknown(msg);

		return;
	}

	BYTE ship = msg->msg[1];
	Player * p = msg->src->getPlayer(getShort(msg->msg, 2));
	if (p == NULL) return;
	Uint16 team = getShort(msg->msg, 4);

	printf("Player changed team and ship: %s\n", p->name);

	p->team = team;
	p->ship = (enum Ship_Types)ship;
}

void handlePlayerBanner		(hostMessage * msg)
{
	/*	Field	Length	Description
		0		1		Type byte
		1		2		Player ident
		3		96		Player banner
	*/

	if (msg->len != 99)
	{
		handleUnknown(msg);

		return;
	}

	Player * p = msg->src->getPlayer(getShort(msg->msg, 1));
	if (p == NULL) return;
	char * banner = msg->msg + 3;

	p->setBanner(banner);
}

void handlePlayerLeaving	(hostMessage * msg)
{
	/*	Field	Length	Description
		0		1		Type byte
		1		2		Player ident
	*/

	if (msg->len != 3)
	{
		handleUnknown(msg);

		return;
	}

	Uint16 ident = getShort(msg->msg, 1);

	Player * p = msg->src->getPlayer(getShort(msg->msg, 1));
	if (p == NULL) return;
	printf("Player leaving: %s\n", p->name);

	msg->src->killPlayer(ident);
}

void handleChat				(hostMessage * msg)
{
	/*	Field	Length	Description
		0		1		Type byte
		1		1		Message type
		2		1		Soundcode
		3		2		Player ident
		5		...\0	Message
	*/

	if (msg->len < 6)
	{
		printf("WARNING: Chat message buffer underrun ignored.\n");
		WriteLog(msg->msg, msg->len);

		return;
	}

	if (msg->msg[msg->len - 1])
	{
		printf("WARNING: Unterminated chat message ignored.\n");
		WriteLog(msg->msg, msg->len);

		return;
	}

	BYTE type = getByte(msg->msg, 1);
	BYTE soundcode = getByte(msg->msg, 2);
	Uint16 ident = getShort(msg->msg, 3);
	Player * p = msg->src->getPlayer(ident);
	if ((p == NULL) && (ident != PROT_ERROR)) return;
	char * text = msg->msg + 5;

	if (!*text) return;

	switch (type)
	{
	case MSG_Arena:
			printf("AR %s\n", text);
		break;
	case MSG_PublicMacro:
			if (p) printf("PM %s> %s\n", p->name, text);
		break;
	case MSG_Public:
			if (p) printf("PU %s> %s\n", p->name, text);
		break;
	case MSG_Team:
			if (p) printf("TE %s> %s\n", p->name, text);
		break;
	case MSG_TeamPrivate:
			if (p) printf("TP %s> %s\n", p->name, text);
		break;
	case MSG_Private:
			if (p) printf("PV %s> %s\n", p->name, text);
		break;
	case MSG_PlayerWarning:
			if (p) printf("PW %s> %s\n", p->name, text);
		break;
	case MSG_RemotePrivate:
			printf("RP %s\n", text);
		break;
	case MSG_ServerError:
			printf("SE %s\n", text);
		break;
	case MSG_Channel:
			printf("CH %s\n", text);
		break;
	default:
		printf("WARNING: Unknown chat message ignored. (%i)\n", type);
		WriteLog(msg->msg, msg->len);
		return;
	};

	if (*text == '!') gotCommand(msg->src, p, text + 1);
}

void handleScoreUpdate		(hostMessage * msg)
{
	/*	Field	Length	Description
		0		1		Type byte
		1		2		Player ident
		3		4		Flag points
		7		4		Kill points
		11		2		Wins
		13		2		Losses
	*/

	if (msg->len != 15)
	{
		handleUnknown(msg);

		return;
	}

	Player * p = msg->src->getPlayer(getShort(msg->msg, 1));
	if (p == NULL) return;
	Uint32 flagPoints = getLong(msg->msg, 3);
	Uint32 killPoints = getLong(msg->msg, 7);
	Uint16 wins = getShort(msg->msg, 11);
	Uint16 losses = getShort(msg->msg, 13);

	printf("Score update: %s\n", p->name);

	p->flagPoints = flagPoints;
	p->killPoints = killPoints;
	p->wins = wins;
	p->losses = losses;
}

void handleWeaponUpdate		(hostMessage * msg)
{
	/*	Field	Length	Description
		0		1		Type byte
		1		1		Rotation
		2		2		Timestamp
		4		2		X pixels
		6		2		Y velocity
		8		2		Player ident
		10		2		X velocity
		12		1		Checksum
		13		1		Togglables
		14		1		Ping
		15		2		Y pixels
		17		2		Bounty
		19		2		Weapon info
	Spectating with ExtraPositionData or *energy
		21		2		Energy
	Spectating with ExtraPositionData
		23		2		S2CLag
		25		2		Timer
		27		4		Item information
	*/

	if ((msg->len != 21) && (msg->len != 23) && (msg->len != 31))
	{
		handleUnknown(msg);

		return;
	}

	stateInfo si;
	weaponInfo wi;
	itemInfo ii;

	bool gotItems, gotEnergy;

	switch (msg->len)
	{
	case 23:
		gotItems = false;
		gotEnergy = true;
		break;
	case 31:
		gotItems = true;
		gotEnergy = true;
		break;
	default:
		gotItems = false;
		gotEnergy = false;
	};

	BYTE direction = getByte(msg->msg, 1);

	Uint32 hiword = msg->src->getHostTime() & 0xFFFF0000;
	Uint32 loword = getShort(msg->msg, 2);
	Uint32 timestamp = msg->src->getLocalTime(hiword | loword);

	Uint16 x = getShort(msg->msg, 4);
	Uint16 yvel = getShort(msg->msg, 6);
	Player * p = msg->src->getPlayer(getShort(msg->msg, 8));
	if (p == NULL) return;
	Uint16 xvel = getShort(msg->msg, 10);
	BYTE checksum = getByte(msg->msg, 12);
	si.n = getByte(msg->msg, 13);
	BYTE ping = getByte(msg->msg, 14);
	Uint16 y = getShort(msg->msg, 15);
	Uint16 bounty = getShort(msg->msg, 17);
	wi.n = getShort(msg->msg, 19);

	Uint16 energy = getShort(msg->msg, 21);

	Uint16 S2CLag = getShort(msg->msg, 23);
	Uint16 timer = getShort(msg->msg, 25);
	ii.n = getLong(msg->msg, 27);

	msg->msg[12] = 0;
	if (checksum != simpleChecksum(msg->msg, 21))
	{
		printf("Position checksum mismatch: %s\n", p->name);

		return;
	}

	if (wi.s.type != PROJ_None)
		msg->src->weaponCount++;

	p->d = direction;
	if (gotEnergy)
		p->energy = energy;
	else
		p->energy = 0;
	p->timer = timer;
	p->S2CLag = S2CLag;
	p->bounty = bounty;

	p->x = x;
	p->y = y;
	p->xvel = xvel;
	p->yvel = yvel;

	p->stealth = si.s.stealth;
	p->cloak = si.s.cloaked;
	p->xradar = si.s.xradar;
	p->awarp = si.s.awarp;
	p->ufo = si.s.ufo;

	if (gotItems)
	{
		p->shields = ii.s.shields;
		p->super = ii.s.super;
		p->burst = ii.s.burst;
		p->repel = ii.s.repel;
		p->thor = ii.s.thor;
		p->brick = ii.s.brick;
		p->decoy = ii.s.decoy;
		p->rocket = ii.s.rocket;
	}
}

void handlePlayerDeath		(hostMessage * msg)
{
	/*	Field	Length	Description
		0		1		Type byte
		1		1		?
		2		2		Killer ident
		4		2		Killee ident
		6		2		Bounty
		8		2		?
	*/

	if (msg->len != 10)
	{
		handleUnknown(msg);

		return;
	}

	Player * killer = msg->src->getPlayer(getShort(msg->msg, 2));
	Player * killee = msg->src->getPlayer(getShort(msg->msg, 4));
	Uint16 bounty = getShort(msg->msg, 6);

	if (killer)
	{
		killer->wins++;
		killer->killPoints += bounty;
	}

	if (killee)
	{
		killee->losses++;
	}
}

void handleSpecPlayer		(hostMessage * msg)
{
	/*	Field	Length	Description
		0		1		Type byte
		1		2		Player ident
	*/

	if (msg->len != 3)
	{
		handleUnknown(msg);

		return;
	}

	Player * p = msg->src->getPlayer(getShort(msg->msg, 1));
	if (p == NULL) return;

	printf("Player spectated: %s\n", p->name);

	p->ship = SHIP_Spectator;
}

void handlePlayerPosition	(hostMessage * msg)
{
	/*	Field	Length	Description
		0		1		Type byte
		1		1		Rotation
		2		2		Time stamp
		4		2		X pixels
		6		1		Ping
		7		1		Bounty
		8		1		Player ident
		9		1		Togglables
		10		2		Y velocity
		12		2		Y pixels
		14		2		X velocity
	Spectating with ExtraPositionData or *energy
		16		2		Energy
	Spectating with ExtraPositionData
		18		2		S2CLag
		20		2		Timer
		22		4		Item information
	*/

	if ((msg->len != 16) && (msg->len != 18) && (msg->len != 26))
	{
		handleUnknown(msg);

		return;
	}

	stateInfo si;
	itemInfo ii;

	bool gotItems, gotEnergy;

	switch (msg->len)
	{
	case 18:
		gotItems = false;
		gotEnergy = true;
		break;
	case 26:
		gotItems = true;
		gotEnergy = true;
		break;
	default:
		gotItems = false;
		gotEnergy = false;
	};

	BYTE direction = getByte(msg->msg, 1);

	Uint32 hiword = msg->src->getHostTime() & 0xFFFF0000;
	Uint32 loword = getShort(msg->msg, 2);
	Uint32 timestamp = msg->src->getLocalTime(hiword | loword);

	Uint16 x = getShort(msg->msg, 4);
	BYTE ping = getByte(msg->msg, 6);
	BYTE bounty = getByte(msg->msg, 7);
	Player * p = msg->src->getPlayer(getByte(msg->msg, 8));
	if (p == NULL) return;
	si.n = getByte(msg->msg, 9);
	Uint16 yvel = getShort(msg->msg, 10);
	Uint16 y = getShort(msg->msg, 12);
	Uint16 xvel = getShort(msg->msg, 14);

	Uint16 energy = getShort(msg->msg, 16);

	Uint16 S2CLag = getShort(msg->msg, 18);
	Uint16 timer = getShort(msg->msg, 20);
	ii.n = getLong(msg->msg, 22);

	p->d = direction;
	if (gotEnergy)
		p->energy = energy;
	else
		p->energy = 0;
	p->timer = timer;
	p->S2CLag = S2CLag;
	p->bounty = bounty;

	p->x = x;
	p->y = y;
	p->xvel = xvel;
	p->yvel = yvel;

	p->stealth = si.s.stealth;
	p->cloak = si.s.cloaked;
	p->xradar = si.s.xradar;
	p->awarp = si.s.awarp;
	p->ufo = si.s.ufo;

	if (gotItems)
	{
		p->shields = ii.s.shields;
		p->super = ii.s.super;
		p->burst = ii.s.burst;
		p->repel = ii.s.repel;
		p->thor = ii.s.thor;
		p->brick = ii.s.brick;
		p->decoy = ii.s.decoy;
		p->rocket = ii.s.rocket;
	}
}

void handleKoTHReset		(hostMessage * msg)
{
	/*	Field	Length	Description
		0		1		Type byte
		1		1		Adding KoTH timer
		2		4		Reward
		6		2		Player ident
	*/

	if (msg->len != 8)
	{
		handleUnknown(msg);

		return;
	}

	bool adding = getByte(msg->msg, 1) != 0;
	Uint32 reward = getLong(msg->msg, 2);
	Uint16 ident = getShort(msg->msg, 6);
	Player * p = msg->src->getPlayer(ident);

	if (adding)
	{
		if (ident != PROT_ERROR)
		{
			// Add to one
			if (p == NULL) return;

			printf("KoTH add: %s\n", p->name);

			p->koth = true;
		}
		else
		{
			// Add to all
			printf("KoTH add: all\n");

			_listnode <Player> * parse = msg->src->playerlist.head;

			while (parse)
			{
				p = parse->item;
				parse = parse->next;

				p->koth = true;
			}

			clientMessage * cm = generateKoTHReset();
			if (cm)
			{
				msg->src->post(cm->msg, cm->len, true);
				delete cm;
			}
		}
	}
	else
	{
		if (ident != PROT_ERROR)
		{
			// Remove from one
			if (p == NULL) return;

			printf("KoTH remove: %s\n", p->name);

			p->koth = false;
		}
		else
		{
			// Remove from all
			printf("KoTH remove: all\n");

			_listnode <Player> * parse = msg->src->playerlist.head;

			while (parse)
			{
				p = parse->item;
				parse = parse->next;

				p->koth = false;
			}
		}
	}
}


//////// Personal ship ////////

void handleIdent			(hostMessage * msg)
{
	/*	Field	Length	Description
		0		1		Type byte
		1		2		My new ident
	*/

	if (msg->len != 3)
	{
		handleUnknown(msg);

		return;
	}

	Uint16 ident = getShort(msg->msg, 1);

	msg->src->Me = ident;
}

void handleSelfPrize		(hostMessage * msg)
{
	/*	Field	Length	Description
		0		1		Type byte
		1		2		Prize count
		3		2		Prize type
	*/

	if (msg->len != 5)
	{
		handleUnknown(msg);

		return;
	}

	Uint16 count = getShort(msg->msg, 1);
	Uint16 prize = getShort(msg->msg, 3);

	printf("I have been prized\n");
}

void handleShipReset		(hostMessage * msg)
{
	/*	Field	Length	Description
		0		1		Type byte
	*/

	if (msg->len != 1)
	{
		handleUnknown(msg);

		return;
	}

	printf("Personal ship reset\n");
	Player * p = msg->src->getPlayer(msg->src->Me);
	if (p == NULL) return;

	p->awarp = false;
	p->bounty = 0;
	p->brick = 0;
	p->burst = 0;
	p->cloak = false;
	p->decoy = 0;
	p->energy = 0;
	p->portal = 0;
	p->repel = 0;
	p->rocket = 0;
	p->shields = false;
	p->stealth = false;
	p->super = false;
	p->thor = 0;
	p->timer = 0;
	p->ufo = false;
	p->xradar = false;
}

void handleChangePosition	(hostMessage * msg)
{
	/*	Field	Length	Description
		0		1		Type byte
		1		2		X tiles
		3		2		Y tiles
	*/

	if (msg->len != 5)
	{
		handleUnknown(msg);

		return;
	}

	Uint16 x = getShort(msg->msg, 1);
	Uint16 y = getShort(msg->msg, 3);

	Player * me = msg->src->getPlayer(msg->src->Me);
	if (me == NULL) return;
	me->x = x << 4;
	me->y = y << 4;
}

void handleToggleUFO		(hostMessage * msg)
{
	/*	Field	Length	Description
		0		1		Type byte
		1		1		Boolean: available
	*/

	if (msg->len != 2)
	{
		handleUnknown(msg);

		return;
	}

	bool ufo = getByte(msg->msg, 1) != 0;

	printf("Personal UFO: %i\n", ufo);

	Player * me = msg->src->getPlayer(msg->src->Me);
	if (me == NULL) return;
	me->ufo = ufo;
}


//////// Bricks ////////

void handleBrickDrop		(hostMessage * msg)
{
	/*	Field	Length	Description
		0		1		Type byte
	The following are repeated until the end of the message
		1		2		X1 tiles
		3		2		Y1 tiles
		5		2		X2 tiles
		7		2		Y2 tiles
		9		2		Team
		11		2		Brick ident (sent more than once)
		13		4		Timestamp
	*/

	if ((msg->len - 1) & 15)
	{
		handleUnknown(msg);

		return;
	}

	Uint32 i = 1;

	while (i + 16 <= msg->len)
	{
		Uint16 ident = getShort(msg->msg, i + 10);
		if (msg->src->brickExists(ident))
			return;

		Sint16 x1 = getShort(msg->msg, i + 0);
		Sint16 y1 = getShort(msg->msg, i + 2);
		Sint16 x2 = getShort(msg->msg, i + 4);
		Sint16 y2 = getShort(msg->msg, i + 6);
		Uint16 team = getShort(msg->msg, i + 8);
		Player * p = msg->src->getPlayer(msg->src->Me);
		if (p == NULL) return;
		Uint32 timestamp = msg->src->getLocalTime(getLong(msg->msg, i + 12));

		printf("Player dropped brick: %s\n", p->name);

		// Trace brick blocks
		Sint16 mx = (Sint16)sgn(x2 - x1);
		Sint16 my = (Sint16)sgn(y2 - y1);
		Sint16 x = x1, y = y1;

		BYTE brickIcon = (team == p->team) ? ssbTeamBrick : ssbEnemyBrick;
		Brick * b;

		if (mx)
		{
			while (x != x2)
			{
				x += mx;

				b = new Brick;
				b->x = x;
				b->y = y;
				b->time = timestamp + msg->src->settings.BrickTime;
				b->team = team;
				b->ident = ident;
				msg->src->brickList.append(b);
				msg->src->map[getLinear(x, y)] = brickIcon;
			}
		}
		else
		{
			while (y != y2)
			{
				y += my;

				b = new Brick;
				b->x = x;
				b->y = y;
				b->time = timestamp + msg->src->settings.BrickTime;
				b->team = team;
				b->ident = ident;
				msg->src->brickList.append(b);
				msg->src->map[getLinear(x, y)] = brickIcon;
			}
		}

		i += 16;
	}
}


//////// Map ////////

void handleMapInfo			(hostMessage * msg)
{
	/*	Field	Length	Description
		0		1		Type byte
		1		16		Map name
		17		4		Map checksum
	*/

	if (msg->len != 21)
	{
		handleUnknown(msg);

		return;
	}

	char * name = msg->msg + 1;
	Uint32 checksum = getLong(msg->msg, 17);

	if (checksum != getFileChecksum(name))
	{
		printf("Downloading map file: %s\n", name);

		clientMessage * cm = generateLevelRequest();
		if (cm)
		{
			msg->src->post(cm->msg, cm->len, false);
			delete cm;
		}
	}
	else
	{
		printf("Loading map file: %s\n", name);

		fileAccess fa;
		if (fa.open(name, true))
		{
			Uint32 len = fa.getLength();
			char * file = fa.readStuff(0, len);
			convertFileToMatrix(file, (char*)msg->src->map, len);
			msg->src->resetFlags();
			delete []file;
		}
	}
}

void handleMapFile			(hostMessage * msg)
{
	/*	Field	Length	Description
		0		1		Type byte
		1		16		Map name
		17		...		Compressed map
	*/

	char * name = msg->msg + 1;
	char * compressed = msg->msg + 17;
	Uint32 length = msg->len - 17;

	if (name[15])
	{
		printf("WARNING: Unterminated downloaded-file name ignored.\n");

		return;
	}

	for (Uint16 i = 0; i < 16; i++)
	{
		char c = name[i];

		if (c == 0)
			break;

		switch (c)
		{
		case '/':
		case '\\':
			printf("WARNING: Invalid downloaded-file name path ignored. (%s)\n", name);
			return;
		default:
			break;
		};

		if ((c < ' ') || (c > '~'))
		{
			printf("WARNING: Invalid downloaded-file name chars ignored.\n");
			return;
		}
	}

	printf("Map file download completed: %s\n", name);

	fileAccess fa;
	_unlink(name);
	if (fa.open(name, false))
	{
		fa.decompress(compressed, length);
		Uint32 len = fa.getLength();
		char * file = fa.readStuff(0, len);
		convertFileToMatrix(file, (char*)msg->src->map, len);
		msg->src->resetFlags();
		delete []file;
	}
}


//////// Security ////////

void handleSecurity			(hostMessage * msg)
{	// Fields contributed by Kavar!'s research
	/*	Field	Length	Description
		0		1		Type byte
		1		4		Green seed
		5		4		Door seed
		9		4		Timestamp (Kavar! mistook it for a seed)
		13		4		Checksum generator key
	*/

	if (msg->len != 17)
	{
		handleUnknown(msg);

		return;
	}

	Uint32 greenSeed   = getLong(msg->msg, 1);
	Uint32 doorSeed    = getLong(msg->msg, 5);
	Uint32 timestamp   = msg->src->getLocalTime(getLong(msg->msg, 9));
	Uint32 checksumKey = getLong(msg->msg, 13);

	msg->src->ps.randomize(greenSeed, timestamp);

	msg->src->S2CSlowTotal += 0;
	msg->src->S2CFastTotal += (Uint16)msg->src->msgRecv;

	clientMessage * cm = generateSecurityChecksum(generateParameterChecksum(checksumKey, (Uint32*)&msg->src->settings),
												  generateEXEChecksum(checksumKey, "subspace.bin"),
												  generateLevelChecksum(checksumKey, (char*)msg->src->map),
												  (Uint16)msg->src->syncPing,
												  (Uint16)msg->src->avgPing,
												  (Uint16)msg->src->lowPing,
												  (Uint16)msg->src->highPing,
												  msg->src->weaponCount,
												  0,
												  (Uint16)msg->src->msgRecv,
												  msg->src->S2CSlowTotal,
												  msg->src->S2CFastTotal,
												  false);
	if (cm)
	{
		msg->src->post(cm->msg, cm->len, true);
		delete cm;
	}

	msg->src->S2CFastCurrent = 0;
	msg->src->S2CSlowCurrent = 0;
}


//////// Powerball ////////

void handleSoccerGoal		(hostMessage * msg)
{
	/*	Field	Length	Description
		0		1		Type byte
		1		2		Team
		3		4		Points
	*/

	if (msg->len != 7)
	{
		handleUnknown(msg);

		return;
	}

	Uint16 team = getShort(msg->msg, 1);
	Uint32 points = getLong(msg->msg, 3);

	printf("Got soccer goal\n");
}

void handleBallPosition		(hostMessage * msg)
{
	/*	Field	Length	Description
		0		1		Type byte
		1		1		Powerball ident
		2		2		X pixels
		4		2		Y pixels
		6		2		X velocity
		8		2		Y velocity
		10		2		Owner ident
		12		4		Timestamp
	*/

	if (msg->len != 16)
	{
		handleUnknown(msg);

		return;
	}

	BYTE ball = getByte(msg->msg, 1);
	Uint16 x = getShort(msg->msg, 2);
	Uint16 y = getShort(msg->msg, 4);
	Uint16 xvel = getShort(msg->msg, 6);
	Uint16 yvel = getShort(msg->msg, 8);
	Player * p = msg->src->getPlayer(getShort(msg->msg, 10));
	Uint32 timestamp = msg->src->getLocalTime(getLong(msg->msg, 12));
}


//////// Login ////////

void handleCustomMessage	(hostMessage * msg)
{
	/*	Field	Length	Description
		0		1		Type byte
		1		...\0	Message
	*/

	char * text = msg->msg + 1;

	if (text[msg->len - 2])
	{
		printf("WARNING: Unterminated custom message ignored.\n");
		WriteLog(msg->msg, msg->len);

		return;
	}

	printf("Custom response: %s\n", text);
}

void handleArenaList		(hostMessage * msg)
{
	/*	Field	Length	Description
		0		1		Type byte
	The following are repeated until the end of the message
		1		...\0	Arena name
		X		2		Arena population
	*/

	Uint32 index = 1,
		   len;

	while ((index + 3) <= msg->len)
	{
		char * name = msg->msg + index;
		len = strlen(name);

		Uint16 population = getShort(msg->msg, index + len + 1);

		printf("Arena: %s [%i]\n", name, population);

		index += len + 3;
	}
}

void handleArenaSettings	(hostMessage * msg)
{
	/*	Field	Length	Description
		0		1428	Settings buffer
	The type byte is used in the security checksums
	*/

	if (msg->len != 1428)
	{
		handleUnknown(msg);

		return;
	}

	MEMCPY(msg->msg, &msg->src->settings, 1428);

	msg->src->ps.setSettings(&msg->src->settings, msg->src->map);
}

void handlePasswordResponse	(hostMessage * msg)
{
	/*	Field	Length	Description
		0		1		Type byte
		1		1		Accept response meaning
		2		4		Server version
		6		4		?
		10		4		EXE checksum
		14		4		?
		18		1		?
		19		1		Boolean: Request registration form
		20		4		?
		24		4		News checksum (0 = no news file)
		28		2		?
		30		2		?
		32		4		?
	*/

	if (msg->len != 36)
	{
		handleUnknown(msg);

		return;
	}

	BYTE meaning = getByte(msg->msg, 1);
	Uint32 version = getLong(msg->msg, 2);
	Uint32 EXEChecksum = getLong(msg->msg, 10);
	bool requestRegForm = getByte(msg->msg, 19) != 0;
	Uint32 newsChecksum = getLong(msg->msg, 24);

	BOT_INFO * botInfo = &msg->src->botInfo;

	if (requestRegForm)
	{
		clientMessage * cm = generateRegForm(botInfo->realName, botInfo->email, botInfo->city, botInfo->state, botInfo->sex, botInfo->age, botInfo->playAtHome, botInfo->playAtWork, botInfo->playAtSchool, botInfo->processor, botInfo->regName);
		if (cm)
		{
			msg->src->post(cm->msg, cm->len, true);
			delete cm;
		}

		printf("Sending registration form\n");
	}

	if (EXEChecksum != getFileChecksum("subspace.bin"))
	{
		printf("WARNING: Super Moderator access is required to remain connected in this zone.\n");
	}

	if (newsChecksum != getFileChecksum("news.txt"))
	{
		clientMessage * cm = generateNewsRequest();
		if (cm)
		{
			msg->src->post(cm->msg, cm->len, true);
			delete cm;
		}

		printf("Downloading latest news.txt\n");
	}

	switch (meaning)
	{
	case LOG_Continue:			printf("Password accepted for %s.\n", msg->src->botInfo.name);
								printf("Sending arena connect request\n");
		{
			msg->src->activateGameProtocol();
			clientMessage * cm = generateArenaLogin(botInfo->initialArena, botInfo->initialShip, botInfo->xres, botInfo->yres);
			if (cm)
			{
				msg->src->post(cm->msg, cm->len, true);
				delete cm;
			}
		}
		break;
	case LOG_NewUser:			printf("Unknown player, continue as new user?\n");
								printf("Creating account for %s\n", msg->src->botInfo.name);
		{
			clientMessage * cm = generatePassword(true, botInfo->name, botInfo->password, botInfo->machineID, botInfo->timeZoneBias, botInfo->permissionID, botInfo->clientVersion);
			if (cm)
			{
				msg->src->post(cm->msg, cm->len, true);
				delete cm;
			}
		}
		break;
	case LOG_InvalidPassword:	printf("Invalid password for specified user.  The name you have chosen is probably in use by another player, try picking a different name.\n");
			msg->src->disconnect(true);
		break;
	case LOG_FullArena:			printf("This arena is currently full, try again later.\n");
								printf("Sending password for %s\n", msg->src->botInfo.name);
		{
			clientMessage * cm = generatePassword(false, botInfo->name, botInfo->password, botInfo->machineID, botInfo->timeZoneBias, botInfo->permissionID, botInfo->clientVersion);
			if (cm)
			{
				msg->src->post(cm->msg, cm->len, true);
				delete cm;
			}
		}
		break;
	case LOG_LockedOut:			printf("You have been locked out of SubSpace, for more information inquire on Web BBS.\n");
			msg->src->disconnect(true);
		break;
	case LOG_NoPermission:		printf("You do not have permission to play in this arena(1), see Web Site for more information.\n");
			msg->src->disconnect(true);
		break;
	case LOG_SpectateOnly:		printf("You only have permission to spectate in this arena.\n");
								printf("Sending arena connect request\n");
		{
			msg->src->activateGameProtocol();
			clientMessage * cm = generateArenaLogin(botInfo->initialArena, botInfo->initialShip, botInfo->xres, botInfo->yres);
			if (cm)
			{
				msg->src->post(cm->msg, cm->len, true);
				delete cm;
			}
		}
		break;
	case LOG_TooManyPoints:		printf("You have too many points to play in this arena, please choose another arena.\n");
			msg->src->disconnect(true);
		break;
	case LOG_SlowConnection:	printf("Your connection appears to be too slow to play in this arena.\n");
			msg->src->disconnect(true);
		break;
	case LOG_NoPermission2:		printf("You do not have permission to play in this arena(2), see Web Site for more information.\n");
			msg->src->disconnect(true);
		break;
	case LOG_NoNewConnections:	printf("The server is currently not accepting new connections.\n");
								printf("Sending password for %s\n", msg->src->botInfo.name);
		{
			clientMessage * cm = generatePassword(false, botInfo->name, botInfo->password, botInfo->machineID, botInfo->timeZoneBias, botInfo->permissionID, botInfo->clientVersion);
			if (cm)
			{
				msg->src->post(cm->msg, cm->len, true);
				delete cm;
			}
		}
		break;
	case LOG_InvalidName:		printf("Invalid user name entered, please pick a different name.\n");
			msg->src->disconnect(true);
		break;
	case LOG_ObsceneName:		printf("Possibly offensive user name entered, please pick a different name.\n");
			msg->src->disconnect(true);
		break;
	case LOG_BillerDown:		printf("NOTICE: Server difficulties; this zone is currently not keeping track of scores.  Your original score will be available later.  However, you are free to play in the zone until we resolve this problem.\n");
								printf("Sending arena connect request\n");
		{
			msg->src->activateGameProtocol();
			clientMessage * cm = generateArenaLogin(botInfo->initialArena, botInfo->initialShip, botInfo->xres, botInfo->yres);
			if (cm)
			{
				msg->src->post(cm->msg, cm->len, true);
				delete cm;
			}
		}
		break;
	case LOG_BusyProcessing:	printf("The server is currently busy processing other login requests, please try again in a few moments.\n");
								printf("Sending password for %s\n", msg->src->botInfo.name);
		{
			clientMessage * cm = generatePassword(false, botInfo->name, botInfo->password, botInfo->machineID, botInfo->timeZoneBias, botInfo->permissionID, botInfo->clientVersion);
			if (cm)
			{
				msg->src->post(cm->msg, cm->len, true);
				delete cm;
			}
		}
		break;
	case LOG_ExperiencedOnly:	printf("This zone is restricted to experienced players only (ie. certain number of game-hours logged).\n");
			msg->src->disconnect(true);
		break;
	case LOG_UsingDemoVersion:	printf("You are currently using the demo version.  Your name and score will not be kept track of.\n");
			msg->src->disconnect(true);
		break;
	case LOG_TooManyDemos:		printf("This arena is currently has(sic) the maximum Demo players allowed, try again later.\n");
			msg->src->disconnect(true);
		break;
	case LOG_ClosedToDemos:		printf("This arena is closed to Demo players.\n");
			msg->src->disconnect(true);
		break;
	case LOG_UnknownResponse:	printf("Unknown response type, please go to Web site for more information and to obtain latest version of the program.\n");
			msg->src->disconnect(true);
		break;
	};
}


//////// Client generators ////////

clientMessage * generateViolation		(Security_Violations code)
{	clientMessage * ret = new clientMessage(2);
	if (ret == NULL) return NULL;
	char * msg = ret->msg;

	/*	Field	Length	Description
		0		1		Type byte
		1		1		Violation code
	*/

	msg[0] = 0x1B;
	msg[1] = code;

	return ret;
}

clientMessage * generateChangeShip		(Ship_Types ship)
{	clientMessage * ret = new clientMessage(2);
	if (ret == NULL) return NULL;
	char * msg = ret->msg;

	/*	Field	Length	Description
		0		1		Type byte
		1		1		Ship type
	*/

	msg[0] = 0x18;
	msg[1] = ship;

	return ret;
}

clientMessage * generateSpectate		(Uint16 player)
{	clientMessage * ret = new clientMessage(3);
	if (ret == NULL) return NULL;
	char * msg = ret->msg;

	/*	Field	Length	Description
		0		1		Type byte
		1		2		Player ident
	*/

	msg[0] = 0x08;
	*(Uint16*)&msg[1] = player;

	return ret;
}

clientMessage * generateChangeTeam		(Uint16 team)
{	clientMessage * ret = new clientMessage(3);
	if (ret == NULL) return NULL;
	char * msg = ret->msg;

	/*	Field	Length	Description
		0		1		Type byte
		1		2		Team
	*/

	msg[0] = 0x0F;
	*(Uint16*)&msg[1] = team;

	return ret;
}

clientMessage * generateChangeBanner	(BYTE * buffer)
{	clientMessage * ret = new clientMessage(97);
	if (ret == NULL) return NULL;
	char * msg = ret->msg;

	/*	Field	Length	Description
		0		1		Type byte
		1		96		Banner
	*/

	msg[0] = 0x19;
	MEMCPY(buffer, msg + 1, 96);

	return ret;
}

clientMessage * generateDeath			(Uint16 player, Uint16 bounty)
{	clientMessage * ret = new clientMessage(5);
	if (ret == NULL) return NULL;
	char * msg = ret->msg;

	/*	Field	Length	Description
		0		1		Type byte
		1		2		Player ident
		3		2		Bounty
	*/

	msg[0] = 0x0F;
	*(Uint16*)&msg[1] = player;
	*(Uint16*)&msg[3] = bounty;

	return ret;
}

clientMessage * generateChat			(Chat_Modes type, Chat_SoundBytes soundcode, Uint16 player, char * text)
{	Uint32 len = limit(strlen(text), 200);
	clientMessage * ret = new clientMessage(6 + len);
	if (ret == NULL) return NULL;
	char * msg = ret->msg;

	/*	Field	Length	Description
		0		1		Type byte
		1		1		Chat type
		2		1		Soundcode
		3		2		Player ident
		5		...\0	Message
	*/

	msg[0] = 0x06;
	msg[1] = type;
	msg[2] = soundcode;
	*(Uint16*)&msg[3] = player;
	MEMCPY(text, msg + 5, len);
	msg[5 + len] = '\0';

	return ret;
}

clientMessage * generateRegForm			(char * name, char * email, char * city, char * state, RegForm_Sex sex, BYTE age, bool playAtHome, bool playAtWork, bool playAtSchool, Uint32 processor, char * regName)
{	clientMessage * ret = new clientMessage(766);
	if (ret == NULL) return NULL;
	char * msg = ret->msg;

	/*	Field	Length	Description
		0		1		Type byte
		1		32		Real name
		33		64		Email
		97		32		City
		129		24		State
		153		1		Sex('M'/'F')
		154		1		Age
		155		1		Connecting from... home
		156		1						   work
		157		1						   school
		158		4		Processor type (586)
		162		1		?
		163		1		?
		164		3		?
	Contributed by Ave-iator
		166		80		Windows registration name (SSC RegName ban)
	Windows NT-based OS's do not send any hardware specs.
		246		520		Hardware specs (DreamSpec HardwareID ban)
	*/

	memset(msg, 0, 766);

	msg[0] = 0x17;
	MEMCPY(name, msg + 1, limit(strlen(name), 31));
	MEMCPY(email, msg + 33, limit(strlen(email), 63));
	MEMCPY(city, msg + 97, limit(strlen(city), 31));
	MEMCPY(state, msg + 129, limit(strlen(state), 23));
	msg[153] = sex;
	msg[154] = age;
	msg[155] = playAtHome ? 1 : 0;
	msg[156] = playAtWork ? 1 : 0;
	msg[157] = playAtSchool ? 1 : 0;
	*(Uint32*)&msg[158] = processor;
	msg[162] = 0;
	msg[163] = 112;
	*(Uint16*)&msg[164] = 2038;
	msg[166] = 0;
	MEMCPY(regName, msg + 166, limit(strlen(regName), 79));

	return ret;
}

clientMessage * generateKoTHReset		()
{	clientMessage * ret = new clientMessage(1);
	if (ret == NULL) return NULL;
	char * msg = ret->msg;

	/*	Field	Length	Description
		0		1		Type byte
	*/

	msg[0] = 0x1E;

	return ret;
}

clientMessage * generateSecurityChecksum(Uint32 parameterChecksum, Uint32 EXEChecksum, Uint32 levelChecksum, Uint16 ping, Uint16 avgPing, Uint16 lowPing, Uint16 highPing, Uint32 weaponCount, Uint16 S2CSlowCurrent, Uint16 S2CFastCurrent, Uint16 S2CSlowTotal, Uint16 S2CFastTotal, bool slowFrame)
{	clientMessage * ret = new clientMessage(40);
	if (ret == NULL) return NULL;
	char * msg = ret->msg;

	/*	Field	Length	Description
		0		1		Type byte
		1		4		Weapon count
		5		4		Parameter checksum
		9		4		EXE checksum
		13		4		Level checksum
		17		4		S2CSlowTotal
		21		4		S2CFastTotal
		25		2		S2CSlowCurrent
		27		2		S2CFastCurrent
		29		2		?
		31		2		Ping
		33		2		Average ping
		35		2		Low ping
		37		2		High ping
		39		1		Boolean: Slow frame detected
	*/

	msg[0] = 0x1A;
	*(Uint32*)&msg[1] = weaponCount;
	*(Uint32*)&msg[5] = parameterChecksum;
	*(Uint32*)&msg[9] = EXEChecksum;
	*(Uint32*)&msg[13] = levelChecksum;
	*(Uint32*)&msg[17] = S2CSlowTotal;
	*(Uint32*)&msg[21] = S2CFastTotal;
	*(Uint16*)&msg[25] = S2CSlowCurrent;
	*(Uint16*)&msg[27] = S2CFastCurrent;
	*(Uint16*)&msg[29] = 0;
	*(Uint16*)&msg[31] = ping;
	*(Uint16*)&msg[33] = avgPing;
	*(Uint16*)&msg[35] = lowPing;
	*(Uint16*)&msg[37] = highPing;
	msg[39] = slowFrame ? 1 : 0;

	return ret;
}

clientMessage * generatePassword		(bool newUser, char * name, char * pass, Uint32 machineID, Uint16 timezoneBias, Uint32 permissionID, Uint16 clientVersion)
{	clientMessage * ret = new clientMessage(101);
	if (ret == NULL) return NULL;
	char * msg = ret->msg;

	/*	Field	Length	Description
		0		1		Type byte
		1		1		Boolean: New user
		2		32		Name
		34		32		Password
		66		4		Machine ident
		70		1		?
		71		2		Timezone bias
		73		2		?
		75		2		Client version
		77		4		?
		81		4		?
		85		4		Permission ident
		89		12		?
	*/

	msg[0] = 0x09;
	msg[1] = newUser ? 1 : 0;
	memset(msg + 2, 0, 32);
	MEMCPY(name, msg + 2, limit(strlen(name), 19));
	memset(msg + 34, 0, 32);
	MEMCPY(pass, msg + 34, limit(strlen(pass), 19));
	*(Uint32*)&msg[66] = machineID;
	msg[70]	= 0;
	*(Uint16*)&msg[71] = timezoneBias;
	*(Uint16*)&msg[73] = 0x6f9d;
	*(Uint16*)&msg[75] = clientVersion;
	*(Uint32*)&msg[77] = 444;
	*(Uint32*)&msg[81] = 555;
	*(Uint32*)&msg[85] = permissionID;
	memset(msg + 89, 0, 12);

	return ret;
}

clientMessage * generatePowerballRequest(Uint32 timestamp, BYTE ball)
{	clientMessage * ret = new clientMessage(6);
	if (ret == NULL) return NULL;
	char * msg = ret->msg;

	/*	Field	Length	Description
		0		1		Type byte
		1		1		Ball ident
		2		4		Timestamp
	*/

	msg[0] = 0x20;
	msg[1] = ball;
	*(Uint32*)&msg[2] = timestamp;

	return ret;
}

clientMessage * generatePowerballUpdate	(Uint32 timestamp, BYTE ball, Uint16 x, Uint16 y, Uint16 xvelocity, Uint16 yvelocity, Uint16 owner)
{	clientMessage * ret = new clientMessage(16);
	if (ret == NULL) return NULL;
	char * msg = ret->msg;

	/*	Field	Length	Description
		0		1		Type byte
		1		1		Ball ident
		2		2		X pixels
		4		2		Y pixels
		6		2		X velocity
		8		2		Y velocity
		10		2		Owner ident
		12		4		Timestamp
	*/

	msg[0] = 0x1F;
	msg[1] = ball;
	*(Uint16*)&msg[2] = x;
	*(Uint16*)&msg[4] = y;
	*(Uint16*)&msg[6] = xvelocity;
	*(Uint16*)&msg[8] = yvelocity;
	*(Uint16*)&msg[10] = owner;
	*(Uint32*)&msg[12] = timestamp;

	return ret;
}

clientMessage * generateFlagRequest		(Uint16 flag)
{	clientMessage * ret = new clientMessage(3);
	if (ret == NULL) return NULL;
	char * msg = ret->msg;

	/*	Field	Length	Description
		0		1		Type byte
		1		2		Flag ident
	*/

	msg[0] = 0x13;
	*(Uint16*)&msg[1] = flag;

	return ret;
}

clientMessage * generateFlagDrop		()
{	clientMessage * ret = new clientMessage(1);
	if (ret == NULL) return NULL;
	char * msg = ret->msg;

	/*	Field	Length	Description
		0		1		Type byte
	*/

	msg[0] = 0x15;

	return ret;
}

clientMessage * generateEXERequest		()
{	clientMessage * ret = new clientMessage(1);
	if (ret == NULL) return NULL;
	char * msg = ret->msg;

	/*	Field	Length	Description
		0		1		Type byte
	*/

	msg[0] = 0x0B;

	return ret;
}

clientMessage * generateLevelRequest	()
{	clientMessage * ret = new clientMessage(1);
	if (ret == NULL) return NULL;
	char * msg = ret->msg;

	/*	Field	Length	Description
		0		1		Type byte
	*/

	msg[0] = 0x0C;

	return ret;
}

clientMessage * generateNewsRequest		()
{	clientMessage * ret = new clientMessage(1);
	if (ret == NULL) return NULL;
	char * msg = ret->msg;

	/*	Field	Length	Description
		0		1		Type byte
	*/

	msg[0] = 0x0D;

	return ret;
}

clientMessage * generateArenaLogin		(char * arena, Ship_Types ship, Uint16 xres, Uint16 yres)
{	clientMessage * ret = new clientMessage(26);
	if (ret == NULL) return NULL;
	char * msg = ret->msg;

	/*	Field	Length	Description
		0		1		Type byte
		1		1		Ship type
		2		2		?
		4		2		X resolution
		6		2		Y resolution
		8		2		Main arena number
		10		16		Arena name
	*/

	memset(msg + 10, 0, 16);

	msg[0] = 0x01;
	msg[1] = ship;
	*(Uint16*)&msg[2] = 1;
	*(Uint16*)&msg[4] = xres;
	*(Uint16*)&msg[6] = yres;
	if (strlen(arena) == 0)
	{
		// Random main arena
		*(Uint16*)&msg[8] = 0xFFFF;
	}
	else
	{
		if (*arena == '@')
		{
			// Specific main arena
			*(Uint16*)&msg[8] = atoi(arena + 1);
		}
		else
		{
			// Private arena
			*(Uint16*)&msg[8] = 0xFFFD;
			MEMCPY(arena, msg + 10, limit(strlen(arena), 15));
		}
	}

	return ret;
}

clientMessage * generateArenaLeave		()
{	clientMessage * ret = new clientMessage(1);
	if (ret == NULL) return NULL;
	char * msg = ret->msg;

	/*	Field	Length	Description
		0		1		Type byte
	*/

	msg[0] = 0x02;

	return ret;
}

clientMessage * generateAttachRequest	(Uint16 player)
{	clientMessage * ret = new clientMessage(3);
	if (ret == NULL) return NULL;
	char * msg = ret->msg;

	/*	Field	Length	Description
		0		1		Type byte
		1		2		Player ident
	*/

	msg[0] = 0x10;
	*(Uint16*)&msg[1] = player;

	return ret;
}

clientMessage * generateBrickDrop		(Uint16 xtile, Uint16 ytile)
{	clientMessage * ret = new clientMessage(5);
	if (ret == NULL) return NULL;
	char * msg = ret->msg;

	/*	Field	Length	Description
		0		1		Type byte
		1		2		X tiles
		3		2		Y tiles
	*/

	msg[0] = 0x1C;
	*(Uint16*)&msg[1] = xtile;
	*(Uint16*)&msg[3] = ytile;

	return ret;
}

clientMessage * generatePosition		(Uint32 timestamp, BYTE direction, Uint16 x, Uint16 y, Uint16 xvelocity, Uint16 yvelocity, bool stealth, bool cloaked, bool xradar, bool awarp, bool flash, bool safety, bool ufo, Uint16 bounty, Uint16 energy, Projectile_Types type, BYTE level, bool bouncing, bool emp, bool bomb, BYTE shrap, bool secondary, Uint16 timer, Uint16 S2CLag, bool shields, bool super, BYTE burst, BYTE repel, BYTE thor, BYTE brick, BYTE decoy, BYTE rocket, BYTE portal)
{	clientMessage * ret = new clientMessage(32);
	if (ret == NULL) return NULL;
	char * msg = ret->msg;

	/*	Field	Length	Description
	This is the extended version of the c2s position update
		0		1		Type byte
		1		1		Direction
		2		4		Timestamp
		6		2		X velocity
		8		2		Y pixels
		10		1		Checksum
		11		1		Togglables
		12		2		X pixels
		14		2		Y velocity
		16		2		Bounty
		18		2		Energy
		20		2		Weapon info
		22		2		Energy
		24		2		S2C latency
		26		2		Timer
		28		4		Item information
	*/

	stateInfo si;
	si.s.awarp	 = awarp;
	si.s.cloaked = cloaked;
	si.s.flash	 = flash;
	si.s.pack	 = 0;
	si.s.safety	 = safety;
	si.s.stealth = stealth;
	si.s.ufo	 = ufo;
	si.s.xradar	 = xradar;

	weaponInfo wi;
	wi.s.bouncing	= bouncing;
	wi.s.emp		= emp;
	wi.s.fireType	= secondary;
	wi.s.isBomb		= bomb;
	wi.s.level		= level;
	wi.s.shrapCount = shrap;
	wi.s.type		= type;

	itemInfo ii;
	ii.s.brick	 = brick;
	ii.s.burst	 = burst;
	ii.s.decoy	 = decoy;
	ii.s.pack	 = 0;
	ii.s.portal	 = portal;
	ii.s.repel	 = repel;
	ii.s.rocket	 = rocket;
	ii.s.shields = shields;
	ii.s.super	 = super;
	ii.s.thor	 = thor;

	msg[0] = 0x03;
	msg[1] = direction;
	*(Uint32*)&msg[2] = timestamp;
	*(Uint16*)&msg[6] = xvelocity;
	*(Uint16*)&msg[8] = y;
	msg[10] = 0;
	msg[11] = si.n;
	*(Uint16*)&msg[12] = x;
	*(Uint16*)&msg[14] = yvelocity;
	*(Uint16*)&msg[16] = bounty;
	*(Uint16*)&msg[18] = energy;
	*(Uint16*)&msg[20] = wi.n;

	*(Uint16*)&msg[22] = energy;
	*(Uint16*)&msg[24] = S2CLag;
	*(Uint16*)&msg[26] = timer;
	*(Uint32*)&msg[28] = ii.n;

	msg[10] = simpleChecksum(msg, 22);

	return ret;
}

clientMessage * generatePosition		(Uint32 timestamp, BYTE direction, Uint16 x, Uint16 y, Uint16 xvelocity, Uint16 yvelocity, 
										 bool stealth, bool cloaked, bool xradar, bool awarp, bool flash, bool safety, bool ufo,
										 Uint16 bounty, Uint16 energy, Projectile_Types type, BYTE level, bool bouncing, bool emp,
										 bool bomb, BYTE shrap, bool secondary)
{	clientMessage * ret = new clientMessage(22);
	if (ret == NULL) return NULL;
	char * msg = ret->msg;

	/*	Field	Length	Description
	This is the basic version of the c2s position update
		0		1		Type byte
		1		1		Direction
		2		4		Timestamp
		6		2		X velocity
		8		2		Y pixels
		10		1		Checksum
		11		1		Togglables
		12		2		X pixels
		14		2		Y velocity
		16		2		Bounty
		18		2		Energy
		20		2		Weapon info
	*/

	stateInfo si;
	si.s.awarp	 = awarp;
	si.s.cloaked = cloaked;
	si.s.flash	 = flash;
	si.s.pack	 = 0;
	si.s.safety	 = safety;
	si.s.stealth = stealth;
	si.s.ufo	 = ufo;
	si.s.xradar	 = xradar;

	weaponInfo wi;
	wi.s.bouncing	= bouncing;
	wi.s.emp		= emp;
	wi.s.fireType	= secondary;
	wi.s.isBomb		= bomb;
	wi.s.level		= level;
	wi.s.shrapCount = shrap;
	wi.s.type		= type;

	msg[0] = 0x03;
	msg[1] = direction;
	*(Uint32*)&msg[2] = timestamp;
	*(Uint16*)&msg[6] = xvelocity;
	*(Uint16*)&msg[8] = y;
	msg[10] = 0;
	msg[11] = si.n;
	*(Uint16*)&msg[12] = x;
	*(Uint16*)&msg[14] = yvelocity;
	*(Uint16*)&msg[16] = bounty;
	*(Uint16*)&msg[18] = energy;
	*(Uint16*)&msg[20] = wi.n;

	msg[10] = simpleChecksum(msg, 22);

	return ret;
}

clientMessage * generateFileTransfer	(char * fileName, char * buffer, Uint32 len)
{	clientMessage * ret = new clientMessage(17 + len);
	if (ret == NULL) return NULL;
	char * msg = ret->msg;

	/*	Field	Length	Description
	This is the basic version of the c2s position update
		0		1		Type byte
		1		16		File name
		17		...		Compressed file
	*/

	msg[0] = 0x16;
	memset(msg + 1, 0, 16);
	MEMCPY(fileName, msg + 1, limit(strlen(fileName), 15));
	MEMCPY(buffer, msg + 17, len);

	return ret;
}
