/*
	Map file conversion and related constants by Catid@pacbell.net
*/

#ifndef MAP_H
#define MAP_H

#include "datatypes.h"
#include "prng.h"
#include "settings.h"


#define TILE_MAX_X 0x400
#define TILE_MAX_Y 0x400

#define TILE_MAX_LINEAR 0x100000

#define PIXEL_MAX_X 0x4000
#define PIXEL_MAX_Y 0x4000

Uint32 getLinear(Uint32 x, Uint32 y);									// Convert rectangular coords to linear coords

struct tileData
{
	Uint16 x, y;														// Coordinates
	BYTE type;															// Type
};

tileData makeTileData(Uint32 raw);										// File data is in blocks of 32 bits
Uint32 makeTileData(Uint16 x, Uint16 y, BYTE type);						// Produce just such a block

void convertFileToMatrix(char * fileData, char * mapData, Uint32 len);	// Convert from disk to memory format

Uint32 getMapSize(char * mapData);										// Retrieve the length of the map on disk
void convertMatrixToFile(char * mapData, char * _fileData);				// Convert the map to disk format

/* VIE map format

 * X	= WORD((N >> 12) & 0xFFF);	// This puts actual map dimensions at 4095x4095 possible tiles.
 * Y	= WORD(N & 0xFFF);
 * TILE = BYTE(v >> 24);

 * 0		= No tile
 * 1-19		= Normal tiles
 * 20		= Border
 * 21-161	= Normal tiles
 * 162-165	= Vertical doors
 * 166-169	= Horizontal doors
 * 170		= Flag
 * 171		= Safe zone
 * 172		= Goal area
 * 173-175	= Fly over tiles
 * 176-191	= Fly under tiles

 * Warning: Deviating from this format may invalidate the security checksum.

 */

enum MapTileFormat
{
	vieNoTile,			// These are VIE constants

	vieNormalStart		= 1,	// Maps do not place the border for you.
	vieBorder			= 20,	// Borders are included in security checksum
	vieNormalEnd		= 161,	// by a border graphic, and make it impassable.

	vieVDoorStart		= 162,
	vieVDoorEnd			= 165,

	vieHDoorStart		= 166,
	vieHDoorEnd			= 169,

	vieTurfFlag			= 170,

	vieSafeZone			= 171,

	vieGoalArea			= 172,

	vieFlyOverStart		= 173,
	vieFlyOverEnd		= 175,
	vieFlyUnderStart	= 176,
	vieFlyUnderEnd		= 190,

	vieAsteroidStart	= 216,
	vieAsteroidEnd		= 218,

	vieStation			= 219,

	vieWormhole			= 220,

	ssbTeamBrick		= 221,		// These are internal
	ssbEnemyBrick		= 222,

	ssbTeamGoal			= 223,
	ssbEnemyGoal		= 224,

	ssbTeamFlag			= 225,
	ssbEnemyFlag		= 226,

	ssbPrize			= 227,
};


// Prizes

const Uint32 cosmicPrize = 0x10624DD3;

struct prize
{
	Uint16 x, y;
	Sint16 type;
	Uint32 time;
};

class prizeSystem
{	// Obscenely huge thanks go to Kavar! for the synch. system
	Uint32 rlevel;
	SS_HEAVY_PRNG prng;

	Uint32 lastTime;
	_linkedlist <prize> list;

	BYTE * map;
	Uint32 prizeWeightTotal;
	Uint16 prizeFactor;
	Uint16 prizeDelay;
	Uint16 prizeHideCount;
	Uint16 prizeUpgradeVirtual;
	Uint16 prizeMinimumVirtual;
	Uint16 prizeMaxExist;
	Uint16 prizeMinExist;
	Uint16 prizeNegativeFactor;
	prizeSettings prizeWeights;

	Uint32 getTime();

	Uint32 getNumberOfGreens(Uint16 population);
	void decayGreens();
	prize * createGreen(Uint16 population);

public:
	prizeSystem();

	void setSettings(arenaSettings * s, BYTE * m);
	void randomize(Uint32 seed, Uint32 time);
	void doPrizes(Uint16 population);
	void clearList();
};


// Doors

#endif	// MAP_H
